#!/usr/bin/env python3

import importlib.machinery
import importlib.resources
import importlib.util
import os
import sqlite3
import sys
from dataclasses import dataclass
from io import BytesIO, StringIO
from pathlib import Path
from typing import Callable, Optional, Union

import pandas as pd
import tzlocal
from typedstream.stream import TypedStreamReader

import ica.contact as contact
from ica.exceptions import (
    ConversationNotFoundError,
    DateRangeInvalidError,
    FormatNotSupportedError,
)

# In order to interpolate the user-specified list of chat identifiers into the
# SQL queries, we must join the list into a string delimited by a common
# symbol, then perform a LIKE comparison on this string within the SQL query;
# this is because named SQL parameters only support string values, rather than
# variable-length sequences
CHAT_IDENTIFIER_DELIMITER = "|"


# The path to the database file for the macOS Messages application
DB_PATH = Path.home() / "Library" / "Messages" / "chat.db"


# The formats supported for file output, represented as a dictionary where each
# key is the format name and the value is the file extension corresponding to
# that format
SUPPORTED_OUTPUT_FORMAT_MAP = {"csv": "csv", "excel": "xlsx", "markdown": "md"}


@dataclass
class DataFrameNamespace:
    """
    The namespace containing the relevant dataframes for the specified user in
    the chat database
    """

    messages: pd.DataFrame
    attachments: pd.DataFrame


def get_chat_identifier_str(chat_identifiers: list[str]) -> str:
    """
    Join the list of chat identifiers into a delimited string; in order for the
    SQL comparison to function properly, this string must also start and end
    with the delimiter symbol
    """
    return "{start}{joined}{end}".format(
        start=CHAT_IDENTIFIER_DELIMITER,
        joined=CHAT_IDENTIFIER_DELIMITER.join(chat_identifiers),
        end=CHAT_IDENTIFIER_DELIMITER,
    )


def decode_message_attributedbody(data: bytes) -> str:
    """
    The textual contents of some messages are encoded in a special
    attributedBody column on the message row; this attributedBody value is
    in Apple's proprietary typedstream format, but can be parsed with the
    pytypedstream package (<https://pypi.org/project/pytypedstream/>)
    """
    if data:
        for event in TypedStreamReader.from_data(data):
            # The first bytes object is the one we want; it should be safe to
            # convert back to UTF-8
            if type(event) is bytes:
                return event.decode("utf-8")
    return ""


def pipe_lambda(
    df_lambda: Callable[[pd.DataFrame], pd.DataFrame],
) -> Callable[[pd.DataFrame], pd.DataFrame]:
    """
    [DEPRECATED]: Previously, If you were to pass a lambda to the DataFrame
    pipe() method, the type of the dataframe parameter and the return type of
    the lambda would both be `Any`; this broke type-safety for the lambda itself
    and for subsequent chained methods, but to work around this, we introduced a
    wrapper function which, when wrapped around the lambda, can enforce the
    proper type within mypy; however, recent versions of mypy and pandas no
    longer exhibit this behavior, hence we are deprecating this function
    """
    return df_lambda


def assign_lambda(
    df_lambda: Callable[[pd.DataFrame], pd.Series],
) -> Callable[[pd.DataFrame], pd.Series]:
    """
    [DEPRECATED]: Previously, if you were to pass a lambda to the DataFrame
    assign() method, the type of the dataframe parameter and the return type of
    the lambda would both be `Any`; this broke type-safety for the lambda itself
    and for subsequent chained methods, but to work around this, we introduced a
    wrapper function which, when wrapped around the lambda, can enforce the
    proper type within mypy; however, recent versions of mypy and pandas no
    longer exhibit this behavior, hence we are deprecating this function
    """
    return df_lambda


def get_messages_dataframe(
    connection: sqlite3.Connection,
    chat_identifiers: list[str],
    timezone: Optional[str] = None,
) -> pd.DataFrame:
    """
    Return a pandas dataframe representing all messages in a particular
    conversation (identified by the given phone number or email address)
    """
    # If no IANA timezone name is specified, default to the name of the system's
    # local timezone
    if not timezone:
        timezone = tzlocal.get_localzone().key
    return (
        pd.read_sql_query(
            sql=importlib.resources.files(__package__)
            .joinpath(os.path.join("queries", "messages.sql"))
            .read_text(),
            con=connection,
            params={
                "chat_identifiers": get_chat_identifier_str(chat_identifiers),
                "chat_identifier_delimiter": CHAT_IDENTIFIER_DELIMITER,
            },
            parse_dates={"datetime": "ISO8601"},
        )
        # SQL provides each date/time as a Unix timestamp (which is implicitly
        # UTC), but the timestamp is timezone-naive when parsed by pandas; so
        # first, we must add the missing timezone information, then we must
        # convert the datetime to the specified timezone
        .assign(
            datetime=lambda df: df["datetime"]
            .dt.tz_localize("UTC")
            .dt.tz_convert(timezone)
        )
        # Decode any 'attributedBody' values and merge them into the 'text'
        # column
        .assign(
            text=lambda df: df["text"].fillna(
                df["attributedBody"].apply(decode_message_attributedbody)
            )
        )
        # Remove 'attributedBody' column now that it has been merged into the
        # 'text' column
        .drop("attributedBody", axis="columns")
        # Use a regex-based heuristic to determine which messages are reactions
        .assign(
            is_reaction=lambda df: df["text"].str.match(
                r"^(Loved|Liked|Disliked|Laughed at|Emphasized|Questioned|Reacted)"
                r" (“(.*?)”|an \w+|(.*?) to “(.*?)”)$"
            )
        )
        # Convert 'is_from_me' values from integers to proper booleans
        .assign(is_from_me=lambda df: df["is_from_me"].astype(bool))
    )


def filter_dataframe(
    df: pd.DataFrame,
    from_date: Optional[str] = None,
    to_date: Optional[str] = None,
    from_person: Optional[str] = None,
) -> pd.DataFrame:
    """
    Return a copy of the messages dataframe that has been filtered by the
    user-supplied filters
    """

    # Raise an exception if the 'from' date is after the 'to' date
    if from_date and to_date and pd.Timestamp(from_date) > pd.Timestamp(to_date):
        raise DateRangeInvalidError("Date range is backwards")

    return (
        df.pipe(lambda df: df[df["is_from_me"].eq(True)] if from_person == "me" else df)
        .pipe(
            lambda df: (df[df["is_from_me"].eq(False)] if from_person == "them" else df)
        )
        .pipe(lambda df: df[df["datetime"] >= from_date] if from_date else df)
        .pipe(lambda df: df[df["datetime"] <= to_date] if to_date else df)
    )


def get_attachments_dataframe(
    connection: sqlite3.Connection,
    chat_identifiers: list[str],
    timezone: Optional[str] = None,
) -> pd.DataFrame:
    """
    Return a pandas dataframe representing all attachments in a particular
    conversation (identified by the given phone number)
    """
    return (
        pd.read_sql_query(
            sql=importlib.resources.files(__package__)
            .joinpath(os.path.join("queries", "attachments.sql"))
            .read_text(),
            con=connection,
            params={
                "chat_identifiers": get_chat_identifier_str(chat_identifiers),
                "chat_identifier_delimiter": CHAT_IDENTIFIER_DELIMITER,
            },
            parse_dates={"datetime": "ISO8601"},
        )
        # Expose the date/time of the message alongside each attachment record,
        # for convenience
        .assign(
            datetime=lambda df: df["datetime"]
            .dt.tz_localize("UTC")
            .dt.tz_convert(timezone)
        )
    )


def get_dataframes(
    contact_name: str,
    timezone: Optional[str] = None,
    from_date: Optional[str] = None,
    to_date: Optional[str] = None,
    from_person: Optional[str] = None,
) -> DataFrameNamespace:
    """
    Return all dataframes for a specific macOS Messages conversation
    """
    chat_identifiers = contact.get_chat_identifiers(contact_name)
    with sqlite3.connect(DB_PATH) as connection:
        dfs = DataFrameNamespace(
            messages=get_messages_dataframe(connection, chat_identifiers, timezone),
            attachments=get_attachments_dataframe(
                connection, chat_identifiers, timezone
            ),
        )
        if dfs.messages.empty:
            raise ConversationNotFoundError(
                f'No conversation found for the contact "{contact_name}"'
            )
        dfs.messages = filter_dataframe(
            dfs.messages,
            from_date,
            to_date,
            from_person,
        )
        dfs.attachments = filter_dataframe(
            dfs.attachments,
            from_date,
            to_date,
            from_person,
        )
        return dfs


def prettify_header_name(header_name: Union[str, int]) -> Union[str, int]:
    """
    Format the given header name to be more human-readable (e.g. "foo_bar" =>
    "Foo Bar")
    """
    if header_name and type(header_name) is str:
        return header_name.replace("_", " ").title()
    else:
        return header_name


def infer_format_from_output_file_path(output: Optional[str]) -> Optional[str]:
    """
    Assuming an explicit format was not provided, infer the format from the
    extension of the given output file path
    """
    if not output:
        return None
    ext = Path(output).suffix[1:]
    if ext not in SUPPORTED_OUTPUT_FORMAT_MAP.values():
        return None
    return ext


def make_dataframe_tz_naive(df: pd.DataFrame) -> pd.DataFrame:
    """
    Convert all of the datetime timstamps in the given dataframe to be
    timezone-naive, including all columns and the index
    """
    return df.pipe(
        lambda df: (
            df.set_index(df.index.tz_localize(None))
            if isinstance(df.index, pd.DatetimeIndex)
            else df
        )
    ).assign(
        **{
            col: df[col].dt.tz_localize(None)
            for col in df.select_dtypes(include=["datetime64[ns, UTC]"])
        }
    )


def prepare_df_for_output(
    df: pd.DataFrame, prettify_index: bool = True
) -> pd.DataFrame:
    """
    Prepare the given dataframe for output by prettifying column names,
    stripping timezone details incompatible with Excel, and other normalization
    operations; return the normalized dataframe
    """
    return (
        df.rename(
            # Prettify header column (i.e. textual values in first column)
            index=prettify_header_name if prettify_index else None,
            # Prettify header row (i.e. column names)
            columns=prettify_header_name,
        )
        # Prettify index column name
        .rename_axis(prettify_header_name(df.index.name), axis="index")
        # Make all indices start from 1 instead of 0, but only if the index is
        # the default (rather than a custom column)
        .pipe(lambda df: df.set_index(df.index + 1 if not df.index.name else df.index))
        # Make dataframe timestamps timezone-naive (which is required for
        # exporting to Excel)
        .pipe(lambda df: make_dataframe_tz_naive(df))
    )


def output_results(
    analyzer_df: pd.DataFrame,
    format: Optional[str] = None,
    output: Union[str, StringIO, BytesIO, None] = None,
    prettify_index: bool = True,
) -> None:
    """
    Print the dataframe provided by an analyzer module
    """
    is_default_index = not analyzer_df.index.name
    output_df = prepare_df_for_output(analyzer_df, prettify_index=prettify_index)

    if format and format not in {
        *SUPPORTED_OUTPUT_FORMAT_MAP.keys(),
        *SUPPORTED_OUTPUT_FORMAT_MAP.values(),
    }:
        raise FormatNotSupportedError(
            f'The format "{format}" is not supported for output'
        )

    if not format and type(output) is str:
        format = infer_format_from_output_file_path(output)

    output_not_specified: bool = False
    if not output and format in ("excel", "xlsx"):
        output_not_specified = True
        output = BytesIO()
    elif not output:
        output_not_specified = True
        output = StringIO()

    # Keyword arguments passed to any of the to_* output methods
    output_args: dict = {"index": not is_default_index}

    # Output executed DataFrame to correct format
    if format in ("xlsx", "excel"):
        output_df.to_excel(output, **output_args)
    elif format == "csv":
        output_df.to_csv(output, **output_args)
    elif format in ("md", "markdown"):
        output_df.to_markdown(output, **output_args)
    else:
        (
            output_df
            # When we output the dataframe with to_string(), if the index has a
            # name, it will be displayed on a separate line underneath the line
            # with the column names; this is because space needs to be reserved
            # for the columns axis name; to solve this, we can make the name of
            # the index the name of the columns axis, then remove the name from
            # the index (source: <https://stackoverflow.com/a/43635736/560642>)
            .rename_axis(output_df.index.name, axis="columns")
            .rename_axis(None, axis="index")
            .to_string(output, index=True, line_width=100000)
        )

    # Print output if no output file path was supplied
    if output_not_specified and type(output) is BytesIO:
        sys.stdout.buffer.write(output.getvalue())
    elif output_not_specified and type(output) is StringIO:
        print(output.getvalue(), flush=True)



#!/usr/bin/env python3
import pandas as pd

import ica

# The format to use for all date strings
DATE_FORMAT = "%Y-%m-%d"


def main() -> None:
    """
    Generates a comprehensive breakdown of message totals for every day you and
    the other person have been messaging in the conversation
    """
    cli_args = ica.get_cli_parser().parse_args(namespace=ica.TypedCLIArguments())
    dfs = ica.get_dataframes(
        contact_name=cli_args.contact_name,
        timezone=cli_args.timezone,
        from_date=cli_args.from_date,
        to_date=cli_args.to_date,
        from_person=cli_args.from_person,
    )
    ica.output_results(
        (
            dfs.messages[["text", "is_from_me", "datetime", "is_reaction"]]
            # Count all "text" column values by converting them to integers
            # (always 1), because resampling the DataFrame will remove all
            # non-numeric columns
            .assign(
                text=lambda df: df["text"].apply(pd.to_numeric, errors="coerce").isna()
            )
            .resample("D", on="datetime")
            .sum()
            .rename_axis("date", axis="index")
            # Filter out any rows for dates where neither person sent a message
            .pipe(lambda df: df[df["text"] != 0])
            # Add a column for the by-day number of messages from the other
            # person, for convenience (even though it can technically be derived
            # from the existing columns)
            .assign(is_from_them=lambda df: df["text"] - df["is_from_me"])
            # Do not include reaction data for brevity
            .drop(columns=["is_reaction"])
            # Rename columns to be more intuitive
            .rename(
                columns={
                    "text": "#_sent",
                    "is_from_me": "#_sent_by_me",
                    "is_from_them": "#_sent_by_them",
                }
            )
        ),
        format=cli_args.format,
        output=cli_args.output,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import datetime
from typing import Union

import pandas as pd

import ica

# The format to use for all date strings
DATE_FORMAT = "%Y-%m-%d"


def get_first_message_date(dfs: ica.DataFrameNamespace) -> datetime.datetime:
    """
    Retrieve the date of the very first message sent in the conversation
    """
    datestr = str(dfs.messages["datetime"].min().strftime(DATE_FORMAT))
    return datetime.datetime.strptime(datestr, DATE_FORMAT)


def get_dates_between(
    start_date: Union[str, datetime.datetime], end_date: Union[str, datetime.datetime]
) -> list[datetime.datetime]:
    """
    Get all dates between (and including) the given two dates; each date can be
    either of type str or datetime
    """
    return list(pd.date_range(start_date, end_date).strftime(DATE_FORMAT))


def get_sums_by_day(dfs: ica.DataFrameNamespace) -> pd.DataFrame:
    """
    Calculate the text message sums, grouped by date
    """
    return (
        dfs.messages
        # Count all "text" column values by converting them to integers (always
        # 1), because resampling the DataFrame will remove all non-numeric
        # columns
        .assign(text=lambda df: df["text"].apply(pd.to_numeric, errors="coerce").isna())
        .resample("D", on="datetime")
        .sum()
        .rename_axis("date", axis="index")
        .assign(is_from_them=lambda df: df["text"] - df["is_from_me"])
    )


def get_all_message_datestrs(dfs: ica.DataFrameNamespace) -> list[str]:
    """
    Retrieve a list of every date for which at least one message was sent
    """
    sums_by_day = get_sums_by_day(dfs)
    return list(sums_by_day[sums_by_day["text"] > 0].index)


def get_noreply_count(dfs: ica.DataFrameNamespace) -> int:
    """
    Generate a day-by-day breakdown of the number of messages sent where only
    one person sent messages for that day
    """
    sums_by_day = get_sums_by_day(dfs)
    is_from_me = sums_by_day["is_from_me"]
    is_from_them = sums_by_day["is_from_them"]
    return len(
        sums_by_day[
            (is_from_me.eq(0) & is_from_them.gt(0))
            | (is_from_me.gt(0) & is_from_them.eq(0))
        ]
    )


def main() -> None:
    """
    Generate a summary of message and reaction counts, by person and in total,
    as well as other insightful metrics
    """
    cli_args = ica.get_cli_parser().parse_args(namespace=ica.TypedCLIArguments())
    dfs = ica.get_dataframes(
        contact_name=cli_args.contact_name,
        timezone=cli_args.timezone,
        from_date=cli_args.from_date,
        to_date=cli_args.to_date,
        from_person=cli_args.from_person,
    )

    all_datestrs = get_dates_between(
        get_first_message_date(dfs), str(datetime.date.today())
    )
    message_datestrs = get_all_message_datestrs(dfs)

    is_reaction = dfs.messages["is_reaction"]

    totals_map = {
        "messages": len(dfs.messages[is_reaction.eq(False)]),
        "messages_from_me": (
            dfs.messages[is_reaction.eq(False)]["is_from_me"].eq(True).sum()
        ),
        "messages_from_them": (
            dfs.messages[is_reaction.eq(False)]["is_from_me"].eq(False).sum()
        ),
        "reactions": len(dfs.messages[is_reaction.eq(True)]),
        "reactions_from_me": (
            dfs.messages[is_reaction.eq(True)]["is_from_me"].eq(True).sum()
        ),
        "reactions_from_them": (
            dfs.messages[is_reaction.eq(True)]["is_from_me"].eq(False).sum()
        ),
        "days_messaged": len(message_datestrs),
        "days_missed": len(all_datestrs) - len(message_datestrs),
        "days_with_no_reply": get_noreply_count(dfs),
    }
    ica.output_results(
        pd.DataFrame(
            {"metric": tuple(totals_map.keys()), "total": tuple(totals_map.values())},
        ).set_index("metric"),
        format=cli_args.format,
        output=cli_args.output,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import importlib.resources
import json
import re

import pandas as pd

import ica


class MostFrequentEmojisCLIArguments(ica.TypedCLIArguments):
    """Additional CLI arguments specific to the most_frequent_emojis analyzer"""

    result_count: int


def get_emoji_list() -> list[str]:
    """
    Fetch a list of the most popular emoji on the Web to use as a basis for
    computing this metric
    """
    return json.loads(
        importlib.resources.files(__package__).joinpath("emojis.json").read_text()
    )


# Output the occurrences of specific emojis
def main() -> None:
    """
    Generates count data for the top 10 most frequently used emojis across the
    entire conversation
    """
    cli_parser = ica.get_cli_parser()
    cli_parser.add_argument(
        "--result-count",
        type=int,
        default=10,
        help="the number of emoji results to rank",
    )
    cli_args = cli_parser.parse_args(namespace=MostFrequentEmojisCLIArguments())
    dfs = ica.get_dataframes(
        contact_name=cli_args.contact_name,
        timezone=cli_args.timezone,
        from_date=cli_args.from_date,
        to_date=cli_args.to_date,
        from_person=cli_args.from_person,
    )

    emojis = get_emoji_list()
    text = dfs.messages[dfs.messages["is_reaction"].eq(False)].get("text")
    emoji_patt = re.compile(
        f"({'|'.join(re.escape(emoji) for emoji in reversed(emojis))})"
    )
    matches = (
        text.str.findall(emoji_patt).apply(pd.Series).stack().reset_index(drop=True)
    )
    ica.output_results(
        (
            pd.DataFrame({"emoji": matches.value_counts()})
            .rename({"emoji": "count"}, axis="columns")
            .rename_axis("emoji", axis="index")
            .head(cli_args.result_count)
        ),
        format=cli_args.format,
        output=cli_args.output,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import re
import typing

import pandas as pd

import ica


class CountPhrasesArgumentParser(ica.TypedCLIArguments):
    """Additional CLI arguments specific to the count_phrases analyzer"""

    phrases: list[str]
    use_regex: bool
    case_sensitive: bool


# class CountPhrasesArgumentParser(ica.ICAArgumentParser):
#     pass


def get_phrase_counts(
    messages_df: pd.DataFrame,
    phrases: list[str],
    use_regex: bool = False,
    case_sensitive: bool = False,
) -> typing.Generator[int, None, None]:
    return (
        messages_df["text"]
        .str.count(
            re.escape(phrase) if use_regex else phrase,
            flags=re.IGNORECASE if not case_sensitive else 0,
        )
        .sum()
        for phrase in phrases
    )


def main() -> None:
    cli_parser = ica.get_cli_parser()
    cli_parser.add_argument("phrases", nargs="+", help="one or more phrases to count")
    cli_parser.add_argument(
        "--use-regex",
        "-r",
        action="store_true",
        help="if specified, treats phrases as regular expressions",
    )
    cli_parser.add_argument(
        "--case-sensitive",
        "-s",
        action="store_true",
        help="if specified, treats phrases as case-sensitive",
    )
    cli_args = cli_parser.parse_args(namespace=CountPhrasesArgumentParser())
    dfs = ica.get_dataframes(
        contact_name=cli_args.contact_name,
        timezone=cli_args.timezone,
        from_date=cli_args.from_date,
        to_date=cli_args.to_date,
        from_person=cli_args.from_person,
    )
    filtered_messages = dfs.messages[~dfs.messages["is_reaction"]]

    ica.output_results(
        (
            pd.DataFrame(
                {
                    "phrase": cli_args.phrases,
                    "count": get_phrase_counts(
                        filtered_messages,
                        cli_args.phrases,
                        case_sensitive=cli_args.case_sensitive,
                    ),
                    "count_from_me": get_phrase_counts(
                        filtered_messages[filtered_messages["is_from_me"].eq(True)],
                        cli_args.phrases,
                    ),
                    "count_from_them": get_phrase_counts(
                        filtered_messages[filtered_messages["is_from_me"].eq(False)],
                        cli_args.phrases,
                    ),
                }
            ).set_index("phrase")
        ),
        format=cli_args.format,
        output=cli_args.output,
        prettify_index=False,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3


import pandas as pd

import ica


# Count the number of occurrences of the given regular expression pattern within
# the provided Series
def count_occurrences(series: pd.Series, pattern: str) -> int:
    return series.str.count(pattern).sum()


def main() -> None:
    """
    Generate count data by attachment type, including number of Spotify links
    shared, YouTube videos, Apple Music, etc.
    """
    cli_args = ica.get_cli_parser().parse_args(namespace=ica.TypedCLIArguments())
    dfs = ica.get_dataframes(
        contact_name=cli_args.contact_name,
        timezone=cli_args.timezone,
        from_date=cli_args.from_date,
        to_date=cli_args.to_date,
        from_person=cli_args.from_person,
    )

    is_reaction = dfs.messages["is_reaction"]
    totals_map = {
        "gifs": dfs.attachments["mime_type"].eq("image/gif").sum(),
        "youtube_videos": count_occurrences(
            dfs.messages[is_reaction.eq(False)]["text"],
            r"(https?://(?:www\.)(?:youtube\.com|youtu\.be)/(?:.*?)(?:\s|$))",
        ),
        "apple_music": count_occurrences(
            dfs.messages[is_reaction.eq(False)]["text"],
            r"(https?://(?:music\.apple\.com)/(?:.*?)(?:\s|$))",
        ),
        "spotify": count_occurrences(
            dfs.messages[is_reaction.eq(False)]["text"],
            r"(https?://(?:open\.spotify\.com)/(?:.*?)(?:\s|$))",
        ),
        "audio_messages": (dfs.attachments["filename"].str.endswith(".caf").sum()),
        "audio_files": (dfs.attachments["filename"].str.endswith(".m4a").sum()),
        "recorded_videos": (dfs.attachments["mime_type"].eq("video/quicktime").sum()),
    }
    ica.output_results(
        pd.DataFrame({"type": totals_map.keys(), "total": totals_map.values()})
        .set_index("type")
        .sort_values(by="total", ascending=False),
        format=cli_args.format,
        output=cli_args.output,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# ruff: noqa

from ica.cli import get_cli_args, get_cli_parser, TypedCLIArguments
from ica.core import (
    DataFrameNamespace,
    assign_lambda,
    get_dataframes,
    output_results,
    pipe_lambda,
)
from ica.exceptions import (
    BaseAnalyzerException,
    ContactNotFoundError,
    ConversationNotFoundError,
    DateRangeInvalidError,
    FormatNotSupportedError,
)



#!/usr/bin/env python3


class BaseAnalyzerException(Exception):
    """
    A base exception class for iMessage Conversation Analyzer
    """

    pass


class ContactNotFoundError(BaseAnalyzerException):
    """
    Raised when the specified contact was not found in the Contacts database
    """

    pass


class ConversationNotFoundError(BaseAnalyzerException):
    """
    Raised when no conversation was found for the specified contact
    """

    pass


class DateRangeInvalidError(BaseAnalyzerException):
    """
    Raised when the specified date range is mismatched (e.g. the start date is
    after the end date)
    """

    pass


class FormatNotSupportedError(BaseAnalyzerException):
    """
    Raised when the specified format is not supported by the library
    """

    pass



#!/usr/bin/env python3

import glob
import importlib.resources
import os
import sqlite3
from pathlib import Path
from typing import Optional

import pandas as pd
import phonenumbers

from ica.exceptions import ContactNotFoundError

# The glob pattern matching all AddressBook SQL databases to read from
DB_GLOB = (
    Path.home()
    / "Library"
    / "Application Support"
    / "AddressBook"
    / "Sources"
    / "*"
    / "*.abcddb"
)


# The default region for all phone numbers
DEFAULT_PHONE_NUMBER_REGION = "US"


def normalize_phone_number(phone_number: str) -> Optional[str]:
    """
    Normalize the given phone number to the format required for searching in the
    iMessage database
    """
    if not phone_number:
        return None
    return phonenumbers.format_number(
        phonenumbers.parse(phone_number, region=DEFAULT_PHONE_NUMBER_REGION),
        phonenumbers.PhoneNumberFormat.E164,
    )


def normalize_email_address(email_address: str) -> Optional[str]:
    """
    Normalize the given email address to the format required for searching in
    the iMessage database
    """
    if not email_address:
        return None
    normalized_email_address = email_address.strip()
    return normalized_email_address


def get_chat_identifiers(contact_name: str) -> list[str]:
    """
    Fetch the sequence of chat identifiers for the contact with the given name
    """
    chat_identifiers: set[str] = set()
    # There is a separate AddressBook database file for each "source" of
    # contacts (e.g. On My Mac, iCloud, etc.); we must query each of these
    # databases and combine the separate results into a single result set
    for db_path in glob.iglob(str(DB_GLOB)):
        with sqlite3.connect(db_path) as connection:
            rows = pd.read_sql_query(
                sql=importlib.resources.files(__package__)
                .joinpath(os.path.join("queries", "contact.sql"))
                .read_text(),
                con=connection,
                params={"contact_name": contact_name},
            )
            # Combine the results
            chat_identifiers.update(
                rows["ZFULLNUMBER"]
                .apply(normalize_phone_number)
                # The normalization function will convert empty strings to None
                # so that we can filter them out with dropna()
                .dropna()
            )
            chat_identifiers.update(
                rows["ZADDRESS"]
                .apply(normalize_email_address)
                # The normalization function will convert empty strings to None
                # so that we can filter them out with dropna()
                .dropna()
            )

    # Quit if the contact with the specified name could not be found
    if not len(chat_identifiers):
        raise ContactNotFoundError(f'No contact found with the name "{contact_name}"')

    return sorted(chat_identifiers)



#!/usr/bin/env python3

import pandas as pd

import ica


def convert_bool_to_yesno(value: bool) -> str:
    """
    Convert a boolean value to "Yes"/"No"
    """
    return str(value).replace("True", "Yes").replace("False", "No")


def main() -> None:
    """
    Generates a full, unedited transcript of every message, including
    reactions, between you and the other person (attachment files not included)
    """
    cli_args = ica.get_cli_parser().parse_args(namespace=ica.TypedCLIArguments())
    dfs = ica.get_dataframes(
        contact_name=cli_args.contact_name,
        timezone=cli_args.timezone,
        from_date=cli_args.from_date,
        to_date=cli_args.to_date,
        from_person=cli_args.from_person,
    )
    ica.output_results(
        pd.DataFrame(
            {
                "timestamp": dfs.messages["datetime"],
                # Convert 1/0 to Yes/No
                "is_from_me": dfs.messages["is_from_me"].apply(convert_bool_to_yesno),
                "is_reaction": dfs.messages["is_reaction"].apply(convert_bool_to_yesno),
                # U+FFFC is the object replacement character, which appears as the
                # textual message for every attachment
                "message": dfs.messages["text"].replace(
                    r"\ufffc", "(attachment)", regex=True
                ),
            }
        ),
        format=cli_args.format,
        output=cli_args.output,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import contextlib
import importlib.machinery
import importlib.util
import sys
from pathlib import Path
from typing import Union

from ica.exceptions import BaseAnalyzerException

# A module-level flag that is set to True if the user invokes the CLI via the
# `ica` command, meaning that an `analyzer` argument will need to be specified
# as a required CLI argument; otherwise, if the user executes an analyzer file
# directly (via `python`, `python -m`, or as an executable), then there is no
# need for the additional CLI argument because it's intrinsically represented by
# the file being executed
did_user_invoke_cli_directly = False


class TypedCLIArguments(object):
    """
    A subclass of argparse.Namespace that exposes type information for all CLI
    arguments supported by the program
    """

    analyzer: str
    contact_name: str
    timezone: Union[str, None]
    from_date: Union[str, None]
    to_date: Union[str, None]
    from_person: Union[str, None]
    format: Union[str, None]
    output: Union[str, None]


def get_cli_parser() -> argparse.ArgumentParser:
    """
    Retrieve the instance of the parser used to parse user arguments
    from the command line
    """
    global did_user_invoke_cli_directly
    parser = argparse.ArgumentParser()
    # Only accept an analyzer parameter if the user is interacting with the CLI
    # through the `ica` command
    if did_user_invoke_cli_directly:
        parser.add_argument(
            "analyzer",
            type=lambda p: str(Path(p).expanduser()),
            help="the name of a built-in analyzer, or a path to an analyzer file",
        )
    parser.add_argument(
        "--contact-name",
        "-c",
        required=True,
        help="the full name of a macOS contact whose conversation you want to analyze",
    )
    parser.add_argument(
        "--timezone",
        "-t",
        help="the IANA time zone name to use for dates; "
        "defaults to the system's local time zone",
    )
    parser.add_argument(
        "--from-date",
        help="a start date to filter messages by (inclusive); the format must "
        "be ISO 8601-compliant, e.g. YYYY-MM-DD or YYYY-MM-DDTHH:MM:SS",
    )
    parser.add_argument(
        "--to-date",
        help="an end date to filter messages by (exclusive); the format must be"
        " ISO 8601-compliant, e.g. YYYY-MM-DD or YYYY-MM-DDTHH:MM:SS",
    )
    parser.add_argument(
        "--from-person",
        choices=["me", "them", "both"],
        help="A reference to the person by whom to filter messages; accepted "
        "values can be 'me', 'them', or 'both' (the default)",
    )
    parser.add_argument(
        "--format",
        "-f",
        choices=(
            "csv",
            "md",
            "markdown",
            "xlsx",
            "excel",
        ),
        help="an optional export format to output the analyzer results as",
    )
    parser.add_argument(
        "--output",
        "-o",
        help="the path of the file to export analyzer results to; required when"
        " exporting Excel (xlsx) files",
    )

    return parser


# Retrieve the file path of a module given its module path (e.g.
# "ica.analyzers.message_totals")
def get_file_path_from_module_path(module_name: str) -> str:
    """Construct the file path of a module from its specified module path"""
    spec = importlib.util.find_spec(module_name)
    if spec is None or spec.origin is None or spec.origin == "namespace":
        raise ModuleNotFoundError(module_name)
    return spec.origin


def get_cli_args() -> argparse.Namespace:
    """
    [DEPRECATED]: Parse user arguments from the command line; use
    get_cli_parser().parse_args() instead
    """
    return get_cli_parser().parse_args()


def run_analyzer(analyzer: str) -> None:
    """
    Load the given metric file as a Python module, and return the DataFrame
    provided by its analyze() function
    """
    # Check to see if the provided value is the module name of a built-in
    # analyzer, otherwise the value is an analyzer path
    with contextlib.suppress(ModuleNotFoundError):
        analyzer = get_file_path_from_module_path(f"ica.analyzers.{analyzer}").replace(
            "__init__.py", "__main__.py"
        )
    loader = importlib.machinery.SourceFileLoader("__main__", analyzer)
    spec = importlib.util.spec_from_loader(loader.name, loader)
    if not spec:
        raise ImportError(f"Could not create a module spec for {analyzer}")
    analyzer_module = importlib.util.module_from_spec(spec)
    # Expose package information to dynamically-imported module
    analyzer_module.__package__ = __package__
    loader.exec_module(analyzer_module)


def main() -> None:
    """
    Entry point for the `ica` CLI command
    """
    # Keep track of whether the user invokes the CLI directly via the `ica`
    # command
    global did_user_invoke_cli_directly
    did_user_invoke_cli_directly = True

    # parse_known_args() is slightly different from parse_args() in that the
    # former returns a two-item tuple, where the first item is the Namespace of
    # known arguments, and the second item is a list of any unknown arguments
    cli_args = get_cli_parser().parse_known_args(namespace=TypedCLIArguments)[0]

    try:
        run_analyzer(cli_args.analyzer)
    except BaseAnalyzerException as error:
        # Print the error message without the traceback
        print(error, file=sys.stderr)
        sys.exit(1)
    except KeyboardInterrupt:
        print("")


if __name__ == "__main__":
    main()
